"""Module with persistent Raw classes."""

import socket
from datetime import datetime
import pprint
from typing import List, Tuple, Optional

from astropy.io import fits

from micado.main.DataItemExtra import KEY_SCI
from micado.main.Hardware import Hardware
from micado.main.BaseFrame import BaseFrame
from micado.main.Simulator import SimulatorRun
from micado.main.Template import Template


class RawExtra(BaseFrame):
    """A Raw DataItem."""

    def __init__(self, *args, **kwargs):
        """Set detector from extensions if necessary."""
        super().__init__(*args, **kwargs)

    @classmethod
    def layer_names(cls) -> Tuple[List[str], List[str]]:
        """The names of the layers of this data product."""
        # By default no name
        return [KEY_SCI], []

    @classmethod
    def layers_number(cls) -> Tuple[Optional[int], Optional[int]]:
        """Minimum/maximum number of detectors for each layer."""
        # Overloaded in SkyHCI
        from micado.main.Sky import Sky
        from micado.main.Raw import Raw

        # TODO: Generalize this exception?
        if cls in [Raw, Sky]:
            return 1, 9
        return 9, 9

    @property
    def scopesim_dictionaries(self):
        """Generate ScopeSIM dictionaries."""
        props_hw = {
            name: self.__getattribute__(name)
            for name in self.get_persistent()
            if isinstance(self.__getattribute__(name), (Template, Hardware))
        }

        return [hw.scopesim_dictionary for name, hw in props_hw.items()]

    def make(self):
        """Make this object by simulating it."""
        self.make_simulation()

    def make_simulation(self):
        """Simulate data using ScopeSIM."""
        # ScopeSIM is only imported here, so the other functionality of
        # MicadoWISE can be used also when scopesim is not available.
        # TODO: Provide some graceful error message?
        import scopesim
        import scopesim_templates

        # TODO: Move these downloads elsewhere (or make them redundant by
        #       putting the necessary files on the dataserver with the
        #       corresponding object.
        scopesim.server.download_package("telescopes/ELT")
        scopesim.server.download_package("instrumnents/MICADO")
        scopesim.server.download_package("instrumnents/MAORY")
        scopesim.server.download_package("locations/Armazones")

        # TODO: ensure observing_block has a role

        pprint.pprint(self.scopesim_dictionaries)

        target = self.observing_block.target

        dx, dy = self.template.offsets[self.tpl__expno]

        src_function = getattr(scopesim_templates.micado, target)
        src = src_function()
        src.shift(dx, dy)

        self.simulator = SimulatorRun()
        self.simulator.run_date = datetime.now()
        self.simulator.machine_dns = socket.getfqdn()

        # The optical train of ScopeSIM currently needs these values.
        # They are defined in the Instrument dictionary. However, the
        # instrument is not included when making a Dark, so we have it here.
        # TODO: Fix this !INST.pixel_scale hack.
        # TODO: Also create a proper "Observation" thing.
        observations_dict = {
            # in arcsec / pixel
            "!INST.pixel_scale": 0.004,
            "!INST.plate_scale": 0.2666667,
            "!OBS.airmass": 1.2,
            "!OBS.declination": -30,
            "!OBS.hour_angle": 0,
            "!OBS.pupil_angle": 0,
            # TODO: Ensure that this does not overwrite the actual temperature
            "!ATMO.temperature": 7.0,
        }
        cmd = scopesim.UserCommands(
            yamls=self.scopesim_dictionaries,
            properties=observations_dict,
        )
        opt = scopesim.OpticalTrain(cmd)
        opt.observe(src)
        hdulistlist = opt.readout()
        hdulist = hdulistlist[0]

        # TODO: reenable simulator, maybe pointing the other way?
        # duration = datetime.now() - self.simulator.run_date
        # self.simulator.run_duration = duration.total_seconds()

        # TODO: reenable reading from FITS headers once they meet the standard
        #       (or vice versa)
        # self.load_hdulist(hdulist)

        hdus: List[fits.ImageHDU] = [
            hdu for hdu in hdulist[1:] if isinstance(hdu, fits.ImageHDU)
        ]
        for exti, (extension, hdu) in enumerate(zip(self.extensions, hdus), 1):
            extension.image = hdu.data
            extension.EXTNAME = f"DET{exti}.DATA"
