# -*- coding: utf-8 -*-
"""Tests for Science.py.

TODO: Should these tests be automated?
      Most of this could be copied verbatim from dark_test.py
      Maybe easy once backward chaining is properly working.
"""

import shutil
import tempfile

import numpy

from micado.config.Environment import Env
from micado.hardware.FieldConstantPSF import FieldConstantPSF
from micado.hardware.FilterCurve import FilterCurve
from micado.hardware.Instrument import Instrument
from micado.hardware.PupilWheel import PupilWheel
from micado.hardware.RelayOptics import RelayOptics
from micado.hardware.ShotNoise import ShotNoise
from micado.hardware.SummedExposure import SummedExposure
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.hardware.LinearityCurve import LinearityCurve
from micado.hardware.QuantumEfficiencyCurve import QuantumEfficiencyCurve
from micado.hardware.SurfaceList import SurfaceList
from micado.hardware.NonCommonPathAberration import NonCommonPathAberration
from micado.hardware.Atmosphere import Atmosphere
from micado.hardware.AtmosphericTERCurve import AtmosphericTERCurve
from micado.hardware.AtmosphericDispersion import AtmosphericDispersion
from micado.hardware.Vibration import Vibration
from micado.hardware.Telescope import Telescope
from micado.dataitems.Science import Science
from micado.dataitems.ScienceExtension import ScienceExtension
from micado.templates.MicadoImgObs import MicadoImgObs
from micado.test.functional.dark_test import remove_file_and_links
from micado.util.esokeywords import verify_fitsheaders


def get_filt_inst_ro():
    """Get FilterCurve Instrument and RelayOptics."""
    filt1 = FilterCurve(
        name="blank",
        minimum_throughput=0.000101,
        outer=0.2,
    )
    filt2 = FilterCurve(
        name="Ks",
        minimum_throughput=0.000101,
        outer=0.2,
    )
    pupilwheel = PupilWheel()

    inst = Instrument(
        name="MICADO",
        ocs__pxscale=0.004,
        plate_scale=0.26666666666,
        filtercurve1=filt1,
        filtercurve2=filt2,
        pupilcurve=pupilwheel,
        mode_mirrors=SurfaceList(realfilename="LIST_MICADO_mirrors_wide.dat"),
        static_mirrors=SurfaceList(realfilename="LIST_MICADO_mirrors_static.dat"),
        ncpa=NonCommonPathAberration(
            realfilename="INST_MICADO_wavefront_error_budget.dat"
        ),
    )
    ro = RelayOptics(
        relay_psf=FieldConstantPSF(
            realfilename="PSF_SCAO_ConstPSF_0_5off.fits",
        )
    )
    return filt1, filt2, inst, ro


def get_atmo_telescope():
    """Get Atmosphere and Telescope."""
    atmo = Atmosphere(
        altitude=3060.0,
        longitude=-70.1918,
        latitude=-24.5899,
        temperature=7.0,
        humidity=0.1,
        pressure=0.755,
        pwv=2.5,
        tercurve=AtmosphericTERCurve(realfilename="TER_armazones_default_NIR_IMG.dat"),
        dispersion=AtmosphericDispersion(),
    )
    telescope = Telescope(
        name="ELT",
        surfaces=SurfaceList(realfilename="LIST_mirrors_ELT.tbl"),
        vibration=Vibration(
            fwhm=0.001,
        ),
    )
    return atmo, telescope


def test_science():
    """Test creation of a Raw Science image."""
    Env["use_python_logging"] = True

    filt1, filt2, inst, ro = get_filt_inst_ro()
    atmo, telescope = get_atmo_telescope()

    # Set the dark current and create a DetectorArray.
    dark_current = 100.0
    mydetector = Detector(
        xhw=0.24,
        yhw=0.24,
    )
    mydetectorlist = [mydetector]
    mydarkcurrent = DarkCurrent(value=dark_current)
    myshotnoise = ShotNoise(state=1)
    myexposureaction = SummedExposure()

    mydetectorarray = DetectorArray(
        name="testdetector",
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
        exposure_action=myexposureaction,
        qecurve=QuantumEfficiencyCurve(realfilename="QE_detector_H2RG.dat"),
        linearity=LinearityCurve(realfilename="FPA_linearity.dat"),
    )

    # A template
    science_template = MicadoImgObs(det__dit=1, det__ndit=1)

    # An observingblock
    science_ob = ObservingBlock(
        name="MyScienceTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[science_template],
    )

    # Extensions
    myextensions = [
        ScienceExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
    ]

    science1 = Science(
        detector=mydetectorarray,
        instrument=inst,
        relay_optics=ro,
        atmosphere=atmo,
        telescope=telescope,
        template=science_template,
        observing_block=science_ob,
        extensions=myextensions,
    )
    science1.make_simulation()

    science1.save_fits()
    fn = science1.filename

    science2 = Science(filename=fn)

    assert numpy.all(science2.extensions[0].image == science1.extensions[0].image)
    assert science1.extensions[0].NAXIS1 == science2.extensions[0].NAXIS1
    assert science1.extensions[0].NAXIS2 == science2.extensions[0].NAXIS2
    assert science1.detector == science2.detector
    assert science1.template == science2.template
    assert science1.observing_block == science2.observing_block
    ext1 = science1.extensions[0]
    ext2 = science2.extensions[0]
    assert ext1.detector == ext2.detector
    assert ext1 == ext2
    assert science2 == science1

    fn3 = tempfile.NamedTemporaryFile(suffix=".fits", delete=False).name
    shutil.copy(fn, fn3)

    science3 = Science(filename=fn3)
    science3.unique_filename()

    hash1 = science1.filename.split("_")[-1]
    hash3 = science3.filename.split("_")[-1]
    assert hash1 == hash3, "Bad file hashes"

    verify_fitsheaders(fn)

    remove_file_and_links(science1.filename)
    remove_file_and_links(science3.filename)
    remove_file_and_links(fn3)


def test_science_multi():
    """Test creation of a multi-extension raw image."""
    Env["use_python_logging"] = True
    # Set the dark current and create a DetectorArray.
    dark_current = 100.0
    mydarkcurrent = DarkCurrent(value=dark_current)

    mydetector1 = Detector(id="det1", x_cen=0, xhw=0.24, yhw=0.24)
    mydetector2 = Detector(id="det2", x_cen=400, xhw=0.24, yhw=0.24)

    mydetectorlist = [mydetector1, mydetector2]
    myshotnoise = ShotNoise(state=1)

    mydetectorarray = DetectorArray(
        name="testdetector",
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
        exposure_action=None,
    )

    filt1, filt2, inst, ro = get_filt_inst_ro()

    # A template
    science_template = MicadoImgObs(det__dit=1)

    # An observingblock
    science_ob = ObservingBlock(
        name="MyScienceTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[science_template],
    )

    myscienceextensions = [
        ScienceExtension(detector=mydetector, BITPIX=-64, NAXIS=2)
        for mydetector in mydetectorlist
    ]

    for myext in myscienceextensions:
        myext.image = numpy.ones(
            (myext.NAXIS1, myext.NAXIS2),
            dtype=numpy.float64,
        )

    science1 = Science(
        detector=mydetectorarray,
        instrument=inst,
        relay_optics=ro,
        template=science_template,
        observing_block=science_ob,
        extensions=myscienceextensions,
    )

    from pprint import pprint

    pprint(science1.fits_keywords)
    science1.save_fits()
    fn = science1.filename

    science2 = Science(filename=fn)

    assert science2.detector.dark_current.value == science1.detector.dark_current.value
    assert science2.detector == science1.detector
    assert science2.extensions[0] == science1.extensions[0]
    assert science2.extensions[1] == science1.extensions[1]
    assert science2 == science1

    remove_file_and_links(science1.filename)
