# -*- coding: utf-8 -*-
"""Create ditaa files.

Ditaa files contain ASCII art that can easily be converted to png or svg files.
See https://github.com/stathissideris/ditaa for more information.
"""

from dataclasses import dataclass, field
from typing import List
from enum import Enum


class Color(Enum):
    """Colors from https://personal.sron.nl/~pault/ ."""

    TRANS = "trans"
    NONE = "   "
    BLACK = "cBLK"
    GRAY = "c999"
    RED = "cB56"
    YELLOW = "cCB4"
    BLUE = "c47A"
    ORANGE = "cE73"
    GREEN = "c283"
    CYAN = "c6CE"
    WHITE = "cFFF"


Color.RECIPE = Color.RED
Color.RAW = Color.GRAY
Color.INTERMEDIATE = Color.BLUE
Color.REFERENCE = Color.YELLOW
Color.OTHERPIPELINE = Color.CYAN
Color.FINAL = Color.GREEN


class Art:
    """ASCII Art."""

    color = "trans"
    text = ""

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        return [self.text]

    @property
    def height(self):
        """Height of this Art."""
        return len(self.lines)

    @property
    def width(self):
        """Width of this Art."""
        return len(self.lines[0])

    def __str__(self):
        """ASCII version of this Art."""
        return "".join(s + "\n" for s in self.lines)


@dataclass
class PosArt:
    """ASCII-Art with a certain position (on a canvas)."""

    x: int = 0
    y: int = 0
    art: Art = None
    centered: bool = False

    @property
    def bottom(self):
        """Bottom position."""
        return self.y + (self.art.height // 2 if self.centered else self.art.height)

    @property
    def rightp(self):
        """Right position."""
        return self.x + (self.art.width // 2 if self.centered else self.art.width)

    @property
    def top(self):
        """Top position."""
        return (self.bottom - self.art.height) if self.centered else self.y

    @property
    def leftp(self):
        """Left position."""
        return (self.rightp - self.art.width) if self.centered else self.x

    @property
    def center_horizontal(self):
        """Horizontal center."""
        return self.x if self.centered else (self.rightp + self.x) / 2

    @property
    def center_vertical(self):
        """Vertical center."""
        return self.y if self.centered else (self.bottom + self.y) / 2

    def is_inside(self, x, y):
        """Is the point at x, y inside this Art?

        Edge should not count, unless the weight and height is one
        (a ProcessMini).

        TODO: Fix the off-by-one error that is probably in here.
        """
        in1 = (self.top + 1 < y < self.bottom - 1) and (
            self.leftp + 1 < x < self.rightp - 1
        )
        # ProcessMini is now a *, so don't put a + over it.
        in2 = (
            self.art.width == 1
            and self.art.height == 1
            and self.x == x + 1
            and self.y == y + 1
        )
        return in1 or in2


@dataclass
class Canvas(Art):
    """A Canvas filled with ASCII-Art. It is ASCII-Art itself too."""

    arts: List[PosArt]
    empty: str = " "
    color: str = "trans"

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        lines = [""]
        for pa in self.arts:
            lines = self._add(lines, pa)

        # Sometimes this can form with rounded corners:
        #
        # /-------------\
        # |             |
        # | PSMFPA_CAT  |
        # |c6CE         |
        # /-------------\
        # |             |
        # | PSMCOLD_CAT |
        # |c6CE         |
        # \-------------/
        #
        # The middle line should be fixed to:
        #
        # /-------------\
        # |             |
        # | PSMFPA_CAT  |
        # |c6CE         |
        # +-------------+
        # |             |
        # | PSMCOLD_CAT |
        # |c6CE         |
        # \-------------/

        lines_2 = lines[:1]
        for i, line in enumerate(lines[1:-1]):
            line_prev = lines[i - 0]
            line_next = lines[i + 2]
            line_2 = "".join(
                "+"
                if (c_prev, c, c_next)
                in [
                    ("|", "/", "|"),
                    ("|", "\\", "|"),
                ]
                else c
                for c_prev, c, c_next in zip(line_prev, line, line_next)
            )
            lines_2.append(line_2)

        lines_2 += lines[-1:]

        return lines_2

    def _add(self, lines, pa: PosArt):
        """Add PosArt pa to the lines."""
        # Information cached because it takes quite some time to retrieve.
        pa_top = pa.top
        pa_art_height = pa.art.height
        pa_leftp = pa.leftp
        pa_art_width = pa.art.width
        while len(lines) < pa_top + pa_art_height:
            lines.append(self.empty * len(lines[0]))
        while len(lines[0]) < pa_leftp + pa_art_width:
            lines = [line + self.empty for line in lines]
        for i, l in enumerate(pa.art.lines):
            # TODO: fix bug when pa.y = -3
            l2 = lines[pa.top + i]
            la = "".join(
                cold
                if (cnew in (" ", self.empty)) & (pa.art.color == "trans")
                else cnew
                for cold, cnew in zip(l2[pa_leftp : pa_leftp + pa_art_width], l)
            )
            l3 = l2[:pa_leftp] + la + l2[pa_leftp + pa_art_width :]

            lines[pa_top + i] = l3
        return lines

    def __str__(self):
        """ASCII version of this Art."""
        return "".join(line + "\n" for line in self.lines)


@dataclass
class ProcessMini(Art):
    """Like Process, but mini."""

    color: str = "cB56"
    text: str = "X"

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        return [
            # "+",
            "*",
            # "/--+--\\",
            # "|cB56 |",
            # # ("XXXXXXX" + self.text)[-7:],
            # "\\--+--/",
        ]


@dataclass
class Box(Art):
    """ASCII-Art that is box-shaped."""

    text: str = ""
    color: str = "   "
    rounded: bool = False
    tag: str = ""

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        tcolor = "    " if self.color == "trans" else self.color
        # TODO: Fix hack of removing 2 from PHOTOMREF_CAT2.
        lines_text = [lt.rstrip("2") + " " for lt in self.text.splitlines()]
        lsin = max(len(sin) for sin in lines_text) + 1
        ltag = len(self.tag)
        ltop = "/" + "-" * lsin + "\\" if self.rounded else "+" + "-" * lsin + "+"
        l2 = (
            "|"
            + (
                (f"{{{self.tag}}}" + " " * (lsin - ltag - 2))
                if self.tag
                else " " * lsin
            )
            + "|"
        )
        lmids = [f"| {sin:<{lsin-1}}|" for sin in lines_text]
        l4 = "|" + tcolor + " " * (lsin - len(tcolor)) + "|"
        lbot = "\\" + "-" * lsin + "/" if self.rounded else "+" + "-" * lsin + "+"
        return [ltop, l2] + lmids + [l4, lbot]

    def __str__(self):
        """ASCII version of this Art."""
        return "".join(s + "\n" for s in self.lines)


@dataclass
class Process(Box):
    """A processing step."""

    # red https://personal.sron.nl/~pault/
    color: str = "cB56"
    rounded: bool = True


@dataclass
class Data(Box):
    """Input or Output data."""

    color: str = "cBLK"
    rounded: bool = True


@dataclass
class DataRaw(Data):
    """Raw data used by this pipeline."""

    color: str = "c999"


@dataclass
class DataStatic(Data):
    """Static calibration data. Not produced in any pipeline."""

    # yellow https://personal.sron.nl/~pault/
    color: str = Color.YELLOW.value


@dataclass
class DataIntermediate(Data):
    """Data created and used by this Pipeline."""

    # blue https://personal.sron.nl/~pault/
    color: str = "c47A"


@dataclass
class DataInternal(Data):
    """Data used internally by this recipe. Not used atm."""

    # orange https://personal.sron.nl/~pault/
    color: str = "cE73"


@dataclass
class DataExternal(Data):
    """Calibration data produced externally by another pipeline.

    E.g. Persistence, PSF, Distortions.
    """

    # cyan https://personal.sron.nl/~pault/
    color: str = Color.CYAN.value


@dataclass
class DataFinal(Data):
    """Final data produced by our pipelines to be used by scientists."""

    # green https://personal.sron.nl/~pault/
    color: str = "c283"


@dataclass
class Template(Box):
    """A Template."""

    color: str = "cFFF"


@dataclass
class RecipeShort(Box):
    """A short recipe, for association matrix."""

    color: str = "cB56"


@dataclass
class RecipeBox(Box):
    """An ASCII-Art Box that represents a Recipe."""

    width: int = 0
    height: int = 0
    color: str = "trans"
    rounded: bool = False

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        win = self.width - 2
        hin = self.height - 4
        ltop = "/" + "-" * win + "\\" if self.rounded else "+" + "-" * win + "+"
        lname = ":" + self.text + " " * (win - len(self.text)) + "|"
        lmid = "|" + " " * win + "|"
        ltag = (
            "|"
            + (
                (f"{{{self.tag}}}" + " " * (win - len(self.tag) - 2))
                if self.tag
                else " " * win
            )
            + "|"
        )
        lbot = "\\" + "-" * win + "/" if self.rounded else "+" + "-" * win + "+"
        return [ltop, lname] + hin * [lmid] + [ltag, lbot]


@dataclass
class ParamsBox(Box):
    """Process Parameters."""

    text: List[str] = field(default_factory=list)
    color: str = "trans"
    rounded: bool = False

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        width_text = max(len(t) for t in self.text)
        lsin = width_text + 2
        ltop = "/" + "-" * lsin + "\\" if self.rounded else "+" + "-" * lsin + "+"
        lstext = [f"| {t:{width_text}} |" for t in self.text]
        lbot = "\\" + "-" * lsin + "/" if self.rounded else "+" + "-" * lsin + "+"
        return [ltop] + lstext + [lbot]


@dataclass
class Arrow(Art):
    """An Arrow.

    Either horizontal or vertical, with positive or negative length.
    """

    vertical: bool = False
    distance: int = 3
    color: str = "trans"
    connector: str = "9"

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        # TODO: fix this dirty debug hack
        connector = "+" if self.connector in "0123456789" else self.connector
        l2 = self.distance - 2
        if not self.vertical:
            lines = (
                ["+" + "-" * l2 + ">"] if self.distance > 0 else ["<" + "-" * l2 + "+"]
            )
        else:
            if self.distance > 0:
                lines = [connector] + ["|"] * l2 + ["v"]
            else:
                lines = ["^"] + ["|"] * l2 + [connector]
        return lines


@dataclass
class Line(Art):
    """A line. Similar to Arrow, but no head."""

    vertical: bool = False
    distance: int = 3
    color: str = "trans"
    style: str = "|"
    connector: str = "8"

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        # TODO: fix this dirty debug hack
        connector = "+" if self.connector in "0123456789" else self.connector
        style_horizontal = {
            "|": "-",
            ":": "=",
        }
        # TODO: Ensure this will not break when self.distance in [1, 2, 3]
        l2 = self.distance - 3
        if self.distance > 0:
            if not self.vertical:
                lines = [
                    connector + style_horizontal[self.style] + "-" * l2 + connector
                ]
            else:
                lines = [connector, self.style] + ["|"] * l2 + [connector]
        else:
            lines = [connector]
        return lines


class Circle(Art):
    """A Circle."""

    color = "cBLK"

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        return [
            "+-----+",
            "|     |",
            "|{o}  |",
            "|cBLK |",
            "+-----+",
        ]


@dataclass
class RecipeStep(Art):
    """ASCII-Art representing a RecipeStep."""

    # TODO: inherit from Canvas directly
    process: Process = None
    inputs: List[Box] = None
    params: ParamsBox = None
    raw: DataRaw = None
    previous: Process = None
    xprocess: int = 10
    xarrow: int = 20
    ws: int = 30
    color: str = "trans"

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        xprocessstep = self.xprocess + (self.ws - self.process.width) // 2
        posarts = []
        y = 2 if len(self.inputs) else 3
        for inp in self.inputs:
            pa = PosArt(0, y, inp)
            yo = y
            y += inp.height + 1
            arr = Arrow(vertical=False, distance=self.xarrow - inp.width + 1)
            paarr = PosArt(inp.width - 1, yo + 2, arr)
            posarts.append(pa)
            posarts.append(paarr)

        if self.params is not None:
            assert self.params.height > 2
            if self.params.height == 3:
                paparams = PosArt(0, y + 1, self.params)
            elif self.params.height in [4, 5]:
                paparams = PosArt(0, y, self.params)
            else:
                paparams = PosArt(0, y, self.params)
                y += self.params.height - 5
            posarts.append(paparams)
            pararr = Arrow(
                vertical=False, distance=xprocessstep - self.params.width + 1
            )
            popararr = PosArt(self.params.width - 1, y + 2, pararr)
            posarts.append(popararr)

        posarts += [
            PosArt(xprocessstep, y, self.process),
            PosArt(self.xarrow, 0, Arrow(vertical=True, distance=y)),
        ]
        c = Canvas(posarts)
        return c.lines

    @property
    def lines_selfcontained(self):
        """Get selfcontained version of recipestep."""
        xprocessstep = self.xprocess + (self.ws - self.process.width) // 2
        posarts = []
        if self.raw is not None:
            prev = self.raw
        elif self.previous is not None:
            prev = self.previous
        else:
            # TODO: find something better for this, this is incorrect
            prev = self.inputs[0]
            self.inputs.pop(0)

        posarts.append(
            PosArt(xprocessstep + (self.process.width - prev.width) // 2, 0, prev)
        )
        posarts.append(PosArt(0, prev.height - 1, self))
        c = Canvas(posarts)
        return c.lines


@dataclass
class RecipeFull(Art):
    """ASCII-Art representing a full Recipe, with RecipeSteps."""

    template: Template = None
    # TODO: 'raw' is not always a 'raw', rename this property.
    raw: DataRaw = None
    recipe: RecipeBox = None
    steps: List[RecipeStep] = None
    product: Data = None
    color: str = "trans"

    @property
    def lines(self):
        """The ASCII lines of this Art."""
        # TODO: use pydantic dataclasses and do this correction in __post_init__
        wis = (
            [inp.width for step in self.steps for inp in step.inputs]
            + [step.params.width for step in self.steps if step.params is not None]
            + [0]
        )
        wi = max(wis)
        wss = [step.process.width for step in self.steps]
        ws = max(wss)
        if len(self.recipe.text) > ws // 2:
            # The name of the recipe is so large that it will go over the
            # down arrow going into the recipe.
            xprocess = wi + 3
            xarrow = xprocess + len(self.recipe.text)
            wsn = (xarrow - xprocess) * 2
            width_recipe = xarrow - xprocess + ws // 2 + 4
        else:
            # TODO: Create a test with Detrend recipe.
            # Move the processteps one character to the right to make room
            # for the incoming process parameters arrow.
            xprocess = wi + 4
            xarrow = xprocess + ws // 2
            wsn = ws
            width_recipe = wsn + 5
        for step in self.steps:
            step.xprocess = xprocess
            step.xarrow = xarrow
            step.ws = wsn
        self.recipe.width = width_recipe

        y = 0
        pas = []
        if self.template is not None:
            # For very long template names the x-position should be 0.
            xtemplate = max(0, xprocess + (wsn - self.template.width) // 2)
            pas.append(PosArt(xtemplate, y, self.template))
            y += self.template.height
            # This arrow goes through the circle.
            arr = Arrow(vertical=True, distance=9)
            pas.append(PosArt(xarrow, y - 1, arr))
            pas.append(PosArt(xarrow - 3, y + 1, Circle()))
        else:
            circ = Circle()
            pas.append(PosArt(xarrow - 3, y, circ))
            y += circ.height
            arr = Arrow(vertical=True, distance=3)
            pas.append(PosArt(xarrow, y - 1, arr))
        y += arr.height - 1
        pas.append(PosArt(xprocess + (wsn - self.raw.width) // 2, y, self.raw))
        y += self.raw.height
        pas.append(PosArt(xarrow, y - 1, Arrow(vertical=True, distance=3)))
        yrb = y + 1
        y = yrb + 0
        for step in self.steps:
            pas.append(PosArt(0, y, step))
            y = y + step.height - 1

        self.recipe.height = y - yrb + 3
        arr = Arrow(vertical=True, distance=7)
        pas.append(PosArt(xarrow, y, arr))
        y += arr.height
        arr = Arrow(vertical=False, distance=self.recipe.width + wi - xarrow + 2)
        pas.append(PosArt(xarrow, y - 3, arr))
        pas.append(PosArt(self.recipe.width + wi + 2, y - 5, self.product))
        pas.append(PosArt(xarrow - 3, y, Circle()))

        pas.append(PosArt(wi + 1, yrb, self.recipe))

        ca = Canvas(arts=pas)
        return ca.lines


# TODO: use pydantic dataclasses to harmonize these with the above.
@dataclass
class SimpleStepParams:
    """Intermediate class for parameters to a RecipeStep."""

    text: List[str] = field(default_factory=list)


@dataclass
class SimpleStepInput:
    """Intermediate class for input to a RecipeStep."""

    cls: type = None
    text: str = ""


@dataclass
class SimpleStep:
    """Intermediate class for a RecipeStep."""

    process: str = ""
    inputs: List[SimpleStepInput] = None
    params: SimpleStepParams = None
    raw: DataRaw = None
    previous: "SimpleStep" = None


@dataclass
class SimpleRecipe:
    """Intermediate class for a Recipe."""

    template: str = ""
    raw: str = ""
    recipe: str = ""
    steps: List[SimpleStep] = None
    product: str = ""


def testrecipe():
    """Test the generation of ASCII-Art of a Recipe."""
    sr = SimpleRecipe(
        template="McdImgSciObject",
        raw="IMAGE_RAW",
        recipe="mcd_img_detrend",
        steps=[
            SimpleStep(
                process="AddWeightQualityLayer",
                inputs=[SimpleStepInput(cls=DataStatic, text="ILLUM_IMG")],
                params=SimpleStepParams(
                    text=["QI"],
                ),
            ),
            SimpleStep(
                process="CorrectAdditive",
                inputs=[
                    SimpleStepInput(cls=DataIntermediate, text="MASTERDARK"),
                    SimpleStepInput(cls=DataExternal, text="PERSISTENCE"),
                ],
                params=SimpleStepParams(
                    text=["THRESH_LOW", "THRESH_UP_VERYLONG", "A", "B", "C"],
                ),
            ),
            SimpleStep(
                process="UpdateHeader",
                inputs=[],
                params=SimpleStepParams(
                    text=["P", "R"],
                ),
            ),
        ],
        product="DETREND_IMG",
    )

    asteps = [
        RecipeStep(
            process=Process(step.process),
            inputs=[n.cls(n.text) for n in step.inputs],
            params=ParamsBox(text=step.params.text)
            if step.params is not None
            else None,
        )
        for step in sr.steps
    ]

    return RecipeFull(
        template=Template(text=sr.template),
        raw=DataRaw(text=sr.raw),
        steps=asteps,
        recipe=RecipeBox(text=sr.recipe),
        product=DataFinal(sr.product),
    )


if __name__ == "__main__":
    tr = testrecipe()
    print(tr)
